
# Processor types
# (pulled from https://hex-rays.com/products/ida/support/idadoc/618.shtml)
PROCESSOR_8086 = "8086"             # Intel 8086                   ┐
PROCESSOR_80286R = "80286r"         # Intel 80286 real mode        │
PROCESSOR_80286P = "80286p"         # Intel 80286 protected mode   │
PROCESSOR_80386R = "80386r"         # Intel 80386 real mode        │
PROCESSOR_80386P = "80386p"         # Intel 80386 protected mode   │
PROCESSOR_80486R = "80486r"         # Intel 80486 real mode        │
PROCESSOR_80486P = "80486p"         # Intel 80486 protected mode   │
PROCESSOR_80586R = "80586r"         # Intel Pentium & MMX real mode│ (IBM PC family)
PROCESSOR_80586P = "80586p"         # Intel Pentium & MMX prot mode│
PROCESSOR_80686P = "80686p"         # Intel Pentium Pro & MMX      │
PROCESSOR_K62 = "k62"               # AMD K6-2 with 3DNow!         │
PROCESSOR_P2 = "p2"                 # Intel Pentium II             │
PROCESSOR_P3 = "p3"                 # Intel Pentium III            │
PROCESSOR_ATHLON = "athlon"         # AMD K7                       │
PROCESSOR_P4 = "p4"                 # Intel Pentium 4              │
PROCESSOR_METAPC = "metapc"         # Disassemble all IBMPC opcodes┘
PROCESSOR_8085 = "8085"             # Intel 8085                   ┐
PROCESSOR_Z80 = "z80"               # Zilog 80                     │
PROCESSOR_Z180 = "z180"             # Zilog 180                    │ (Zilog 80 family)
PROCESSOR_Z380 = "z380"             # Zilog 380                    │
PROCESSOR_64180 = "64180"           # Hitachi HD64180              │
PROCESSOR_GB = "gb"                 # Gameboy                      ┘
PROCESSOR_Z8 = "z8"                 # Zilog 8                        (Zilog 8 family)
PROCESSOR_860XR = "860xr"           # Intel 860 XR                 ┐ (Intel 860 family)
PROCESSOR_860XP = "860xp"           # Intel 860 XP                 ┘
PROCESSOR_8051 = "8051"             # Intel 8051                     (Intel 51 family)
PROCESSOR_80196 = "80196"           # Intel 80196                  ┐ (Intel 80196 family)
PROCESSOR_80196NP = "80196NP"       # Intel 80196NP, NU            ┘
PROCESSOR_M6502 = "m6502"           # MOS 6502                     ┐ (MOS Technology 65xx family)
PROCESSOR_M65C02 = "m65c02"         # MOS 65c02                    ┘
PROCESSOR_PDP11 = "pdp11"           # DEC PDP/11                     (PDP family)
PROCESSOR_68000 = "68000"           # Motorola MC68000             ┐
PROCESSOR_68010 = "68010"           # Motorola MC68010             │
PROCESSOR_68020 = "68020"           # Motorola MC68020             │
PROCESSOR_68030 = "68030"           # Motorola MC68030             │
PROCESSOR_68040 = "68040"           # Motorola MC68040             │ (Motorola 680x0 family)
PROCESSOR_68330 = "68330"           # Motorola CPU32 (68330)       │
PROCESSOR_68882 = "68882"           # Motorola MC68020 with MC68882│
PROCESSOR_68851 = "68851"           # Motorola MC68020 with MC68851│
PROCESSOR_68020EX = "68020EX"       # Motorola MC68020 with both   │
PROCESSOR_COLFIRE = "colfire"       # Motorola ColdFire            │
PROCESSOR_68K = "68K"               # Motorola MC680x0 all opcodes ┘
PROCESSOR_6800 = "6800"             # Motorola MC68HC00            ┐
PROCESSOR_6801 = "6801"             # Motorola MC68HC01            │
PROCESSOR_6803 = "6803"             # Motorola MC68HC03            │
PROCESSOR_6301 = "6301"             # Hitachi HD 6301              │
PROCESSOR_6303 = "6303"             # Hitachi HD 6303              │
PROCESSOR_6805 = "6805"             # Motorola MC68HC05            │ (Motorola 8bit family)
PROCESSOR_6808 = "6808"             # Motorola MC68HC08            │
PROCESSOR_6809 = "6809"             # Motorola MC68HC09            │
PROCESSOR_6811 = "6811"             # Motorola MC68HC11            ┘
PROCESSOR_6812 = "6812"             # Motorola MC68HC12
PROCESSOR_HCS12 = "hcs12"           # Motorola MC68HCS12
PROCESSOR_6816 = "6816"             # Motorola MC68HC16
PROCESSOR_JAVA = "java"             # java                           (Java family)
PROCESSOR_PPC = "ppc"               # PowerPC big endian           ┐
PROCESSOR_PPCL = "ppcl"             # PowerPC little endian        ┘ (PowerPC family)
PROCESSOR_ARM = "arm"               # ARM little endian            ┐ (ARM family)
PROCESSOR_ARMB = "armb"             # ARM big endian               ┘
PROCESSOR_TMS320C2 = "tms320c2"     # TMS320C2x series             ┐ (TMS 16bit addressing family)
PROCESSOR_TMS320C5 = "tms320c5"     # TMS320C5x series             ┘
PROCESSOR_TMS320C6 = "tms320c6"     # TMS320C6x series             ┐ (TMS VLIW family)
PROCESSOR_TMS320C3 = "tms320c3"     # TMS320C3x series             ┘
PROCESSOR_TMS32054 = "tms32054"     # TMS320C54xx series
PROCESSOR_TMS32055 = "tms32055"     # TMS320C55xx series
PROCESSOR_SH3 = "sh3"               # Renesas SH-3 (little endian) ┐
PROCESSOR_SH3B = "sh3b"             # Renesas SH-3 (big endian)    │ (Renesas SuperH series)
PROCESSOR_SH4 = "sh4"               # Renesas SH-4 (little endian) │
PROCESSOR_SH4B = "sh4b"             # Renesas SH-4 (big endian)    │
PROCESSOR_SH2A = "sh2a"             # Renesas SH-2A (big endian)   ┘
PROCESSOR_AVR = "avr"               # ATMEL AVR                      (ATMEL family)
PROCESSOR_MIPSL = "mipsl"           # MIPS little endian           ┐ (MIPS family:R2000, R3000, R4000,
PROCESSOR_MIPSB = "mipsb"           # MIPS big endian              │ R4200, R4300,  R4400, R4600,
PROCESSOR_MIPSRL = "mipsrl"         # MIPS & RSP little            │ R8000, R10000)
PROCESSOR_MIPSR = "mipsr"           # MIPS & RSP big               │
PROCESSOR_R5900L = "r5900l"         # MIPS R5900 little            ┘
PROCESSOR_R5900R = "r5900r"         # MIPS R5900 big               ┘
PROCESSOR_H8300 = "h8300"           # H8/300x in normal mode       ┐
PROCESSOR_H8300A = "h8300a"         # H8/300x in advanced mode     │ (Hitachi H8 family)
PROCESSOR_H8S300 = "h8s300"         # H8S in normal mode           │
PROCESSOR_H8S300A = "h8s300a"       # H8S in advanced mode         ┘
PROCESSOR_H8500 = "h8500"           # H8/500                         (Hitachi H8/500 family)
PROCESSOR_PIC12CXX = "pic12cxx"     # Microchip PIC 12-bit (12xxx) ┐
PROCESSOR_PIC16CXX = "pic16cxx"     # Microchip PIC 14-bit (16xxx) │ (PIC family)
PROCESSOR_PIC18CXX = "pic18cxx"     # Microchip PIC 16-bit (18xxx) ┘
PROCESSOR_SPARCB = "sparcb"         # SPARC big endian             ┐
PROCESSOR_SPARCL = "sparcl"         # SPARC little endian          ┘ (SPARC family)
PROCESSOR_ALPHAB = "alphab"         # DEC Alpha big endian         ┐
PROCESSOR_ALPHAL = "alphal"         # DEC Alpha little endian      ┘ (ALPHA family)
PROCESSOR_HPPA = "hppa"             # HP PA-RISC big endian          (HP PA-RISC family)
PROCESSOR_DSP56K = "dsp56k"         # Motorola DSP 5600x           ┐
PROCESSOR_DSP561XX = "dsp561xx"     # Motorola DSP 561xx           │ (DSP 56K family)
PROCESSOR_DSP563XX = "dsp563xx"     # Motorola DSP 563xx           │
PROCESSOR_DSP566XX = "dsp566xx"     # Motorola DSP 566xx           ┘
PROCESSOR_C166 = "c166"             # Siemens C166                 ┐
PROCESSOR_C166V1 = "c166v1"         # Siemens C166 v1 family       │
PROCESSOR_C166V2 = "c166v2"         # Siemens C166 v2 family       │ (C166 family)
PROCESSOR_ST10 = "st10"             # SGS-Thomson ST10             │
PROCESSOR_SUPER10 = "super10"       # Super10                      ┘
PROCESSOR_ST20 = "st20"             # SGS-Thomson ST20/C1            (ST20 family)
PROCESSOR_ST20C4 = "st20c4"         # SGS-Thomson ST20/C2-C4
PROCESSOR_ST7 = "st7"               # SGS-Thomson ST7                (ST7 family)
PROCESSOR_IA64L = "ia64l"           # Intel Itanium little endian  ┐
PROCESSOR_IA64B = "ia64b"           # Intel Itanium big endian     ┘ (IA64 family)
PROCESSOR_CLI = "cli"               # Microsoft.Net platform
PROCESSOR_NET = "net"               # Microsoft.Net platform (alias)
PROCESSOR_I960L = "i960l"           # Intel 960 little endian      ┐
PROCESSOR_I960B = "i960b"           # Intel 960 big endian         ┘ (i960 family)
PROCESSOR_F2MC16L = "f2mc16l"       # Fujitsu F2MC-16L             ┐
PROCESSOR_F2MC16LX = "f2mc16lx"     # Fujitsu F2MC-16LX            ┘ (Fujitsu F2MC family)
PROCESSOR_78K0 = "78k0"             # NEC 78k/0
PROCESSOR_78K0S = "78k0s"           # NEC 78k/0s
PROCESSOR_M740 = "m740"             # Mitsubishi 8-bit
PROCESSOR_M7700 = "m7700"           # Mitsubishi 16-bit            ┐
PROCESSOR_M7750 = "m7750"           # Mitsubishi 16-bit            ┘ (Mitsubishi 16-bit family)
PROCESSOR_M32R = "m32r"             # Mitsubishi 32-bit            ┐
PROCESSOR_M32RX = "m32rx"           # Mitsubishi 32-bit extended   ┘ (Mitsubishi 32-bit family)
PROCESSOR_ST9 = "st9"               # STMicroelectronics ST9+
PROCESSOR_FR = "fr"                 # Fujitsu FR family
PROCESSOR_M7900 = "m7900"           # Mitsubishi M7900
PROCESSOR_KR1878 = "kr1878"         # Angstrem KR1878
PROCESSOR_AD218X = "ad218x"         # Analog Devices ADSP
PROCESSOR_OAKDSP = "oakdsp"         # Atmel OAK DSP
PROCESSOR_TRICORE = "tricore"       # Infineon Tricore
PROCESSOR_EBC = "ebc"               # EFI Bytecode
PROCESSOR_MSP430 = "msp430"         # Texas Instruments MSP430
